from __future__ import annotations
from typing import TYPE_CHECKING
from pathlib import Path
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt, Signal
from tabulous._qt._svg import QColoredSVGIcon

if TYPE_CHECKING:
    from tabulous._qt._mainwindow._base import _QtMainWidgetBase

ICON_DIR = Path(__file__).parent.parent / "_icons"


class QMainWindowTitleBar(QtW.QMenuBar):
    """The custom title bar widget for the main window."""

    def __init__(self, parent: _QtMainWidgetBase, icon: QtGui.QIcon) -> None:
        super().__init__(parent)
        self.setFixedHeight(28)
        self._title_icon = self.addMenu(icon, "icon")
        self._title_icon.addAction("Toggle toolbar", parent.toggleToolBarVisibility)
        self._title_icon.addSeparator()
        self._title_icon.addAction("Minimize window", self._minimize_window)
        self._title_icon.addAction("Close", self._close_window)

        self._drag_start = None

        self._corner_buttons = QCornerButtons(self)
        self.setCornerWidget(self._corner_buttons)
        self._corner_buttons.minimizeSignal.connect(self._minimize_window)
        self._corner_buttons.middleSignal.connect(self._toggle_maximize_window)
        self._corner_buttons.closeSignal.connect(self._close_window)

    def setIconColor(self, color: QtGui.QColor):
        self._corner_buttons.setIconColors(color, self.parentWidget().windowState())

    def _minimize_window(self):
        return self.parentWidget().setWindowState(Qt.WindowState.WindowMinimized)

    def _toggle_maximize_window(self):
        parent = self.parentWidget()
        parent.toggleWindowState()
        self._corner_buttons.setMiddleIcon(parent.windowState())

    def _close_window(self):
        return self.parentWidget().close()

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        self._drag_start = event.pos()
        return super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent):
        self._drag_start = None
        return super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, a0: QtGui.QMouseEvent) -> None:
        self._toggle_maximize_window()
        return super().mouseDoubleClickEvent(a0)

    def mouseMoveEvent(self, event: QtGui.QMouseEvent):
        if event.buttons() == Qt.MouseButton.LeftButton:
            if self._drag_start is not None:
                self.parentWidget().move(
                    self.mapToGlobal(event.pos() - self._drag_start)
                )
        else:
            self._drag_start = None
        return super().mouseMoveEvent(event)

    if TYPE_CHECKING:

        def parentWidget(self) -> _QtMainWidgetBase:
            ...


class QCornerButtons(QtW.QToolBar):
    minimizeSignal = Signal()
    middleSignal = Signal()
    closeSignal = Signal()

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self.setMaximumHeight(21)
        self._minimize_icon = QColoredSVGIcon.fromfile(ICON_DIR / "minimize.svg")
        self._maximize_inv_icon = QColoredSVGIcon.fromfile(
            ICON_DIR / "maximize_invert.svg"
        )
        self._maximize_icon = QColoredSVGIcon.fromfile(ICON_DIR / "maximize.svg")
        self._close_icon = QColoredSVGIcon.fromfile(ICON_DIR / "close.svg")

        self._minimize_action = self.addAction(
            "minimize window", self.minimizeSignal.emit
        )
        self._minimize_action.setIcon(self._minimize_icon)
        self._middle_action = self.addAction(
            "toggle window size", self.middleSignal.emit
        )
        self._middle_action.setIcon(self._maximize_icon)
        self._close_action = self.addAction("close window", self.closeSignal.emit)
        self._close_action.setIcon(self._close_icon)

        # hover effect
        close_button = self.widgetForAction(self._close_action)
        close_button.setObjectName("cornerCloseButton")
        close_button.setStyleSheet(
            "#cornerCloseButton:hover { background-color: #FF1818; }"
        )

    def setMiddleIcon(self, state: Qt.WindowState):
        if state == Qt.WindowState.WindowMaximized:
            self._middle_action.setIcon(self._maximize_inv_icon)
        else:
            self._middle_action.setIcon(self._maximize_icon)

    def setIconColors(self, color: QtGui.QColor, state: Qt.WindowState):
        self._minimize_icon = self._minimize_icon.colored(color)
        self._maximize_icon = self._maximize_icon.colored(color)
        self._maximize_inv_icon = self._maximize_inv_icon.colored(color)
        self._close_icon = self._close_icon.colored(color)

        self._minimize_action.setIcon(self._minimize_icon)
        if state == Qt.WindowState.WindowMaximized:
            self._middle_action.setIcon(self._maximize_inv_icon)
        else:
            self._middle_action.setIcon(self._maximize_icon)
        self._close_action.setIcon(self._close_icon)



from __future__ import annotations
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt
import pandas as pd


class QTreeHeaderModel(QtCore.QAbstractTableModel):
    def __init__(
        self,
        df: pd.DataFrame,
        orientation: Qt.Orientation = Qt.Orientation.Horizontal,
        parent: QtCore.QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._df = df
        self._orientation = orientation

    def columnCount(self, parent=None):
        if self._orientation == Qt.Orientation.Horizontal:
            return self._df.columns.shape[0]
        else:  # Vertical
            return self._df.index.nlevels

    def rowCount(self, parent=None):
        if self._orientation == Qt.Orientation.Horizontal:
            return self._df.columns.nlevels
        else:
            return self._df.index.shape[0]

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        row = index.row()
        col = index.column()

        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.ToolTipRole:

            if self._orientation == Qt.Orientation.Horizontal:

                if isinstance(self._df.columns, pd.MultiIndex):
                    return str(self._df.columns[col][row])
                else:
                    return str(self._df.columns[col])

            else:

                if isinstance(self._df.index, pd.MultiIndex):
                    return str(self._df.index[row][col])
                else:
                    return str(self._df.index[row])


class QTreeHeaderView(QtW.QHeaderView):
    def __init__(
        self,
        df: pd.DataFrame,
        orientation: Qt.Orientation = Qt.Orientation.Horizontal,
        parent: QtW.QWidget | None = None,
    ) -> None:
        super().__init__(orientation, parent)
        self.setModel(QTreeHeaderModel(df, orientation))

    def paintSection(
        self, painter: QtGui.QPainter, rect: QtCore.QRect, logicalIndex: int
    ) -> None:
        return super().paintSection(painter, rect, logicalIndex)


class QTableHeaderItem(QtW.QTableWidgetItem):
    def __init__(self, parent, item_parent=None):
        super().__init__(parent)
        self._item_parent = item_parent

    def insertChild(self):
        ...

    def children(self):
        ...


class HeaderModel(QtW.QAbstractItemView):

    ...

    # TODO:
    # df0.columns.levels
    # Out[12]: FrozenList([['b', 'c', 'x'], ['mean', 'std', '']])

    # mi = df0.columns

    # mi.get_indexer
    # Out[14]:
    # <bound method Index.get_indexer of MultiIndex([('b', 'mean'),
    #             ('b',  'std'),
    #             ('c', 'mean'),
    #             ('c',  'std'),
    #             ('x',     '')],
    #            )>

    # mi.get_level_values(0)
    # Out[15]: Index(['b', 'b', 'c', 'c', 'x'], dtype='object')

    # mi.get_level_values(1)
    # Out[16]: Index(['mean', 'std', 'mean', 'std', ''], dtype='object')

    # mi.nlevels
    # Out[17]: 2



from __future__ import annotations
from typing import TYPE_CHECKING
from magicgui import widgets as mwdg
from . import _utils, _dialogs
from tabulous._magicgui import ToggleSwitchSelect

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


def copy_data_tab_separated(viewer: TableViewerBase):
    """Copy cells (tab separated)"""
    _utils.get_table(viewer).native.copyToClipboard(headers=False, sep="\t")


def copy_data_with_header_tab_separated(viewer: TableViewerBase):
    """Copy cells with headers (tab separated)"""
    _utils.get_table(viewer).native.copyToClipboard(headers=True, sep="\t")


def copy_data_comma_separated(viewer: TableViewerBase):
    """Copy cells (comma separated)"""
    _utils.get_table(viewer).native.copyToClipboard(headers=False, sep=",")


def copy_data_with_header_comma_separated(viewer: TableViewerBase):
    """Copy cells with headers (comma separated)"""
    _utils.get_table(viewer).native.copyToClipboard(headers=True, sep=",")


def copy_as_literal(viewer: TableViewerBase):
    """Copy as literal"""
    _utils.get_table(viewer)._qwidget._copy_as_literal()


def copy_as_markdown(viewer: TableViewerBase):
    """Copy as markdown text"""
    _utils.get_table(viewer)._qwidget._copy_as_formated("markdown")


def copy_as_rst_simple(viewer: TableViewerBase):
    """Copy as reStructuredText (rst) simple table"""
    _utils.get_table(viewer)._qwidget._copy_as_formated("rst")


def copy_as_rst_grid(viewer: TableViewerBase):
    """Copy as reStructuredText (rst) grid table"""
    _utils.get_table(viewer)._qwidget._copy_as_formated("grid")


def copy_as_latex(viewer: TableViewerBase):
    """Copy as LaTeX text"""
    _utils.get_table(viewer)._qwidget._copy_as_formated("latex")


def copy_as_html(viewer: TableViewerBase):
    """Copy as HTML text"""
    _utils.get_table(viewer)._qwidget._copy_as_formated("html")


def copy_as_new_table(viewer: TableViewerBase):
    """Copy as new table"""
    _utils.get_table(viewer)._qwidget._copy_as_new_table(type_="table")


def copy_as_new_spreadsheet(viewer: TableViewerBase):
    """Copy as new spreadsheet"""
    _utils.get_table(viewer)._qwidget._copy_as_new_table(type_="spreadsheet")


def select_all(viewer: TableViewerBase):
    """Select all the cells"""
    _utils.get_table(viewer)._qwidget._qtable_view.selectAll()


def cut_data(viewer: TableViewerBase):
    """Cut selected cells"""
    qtable = _utils.get_mutable_table(viewer)._qwidget
    qtable.copyToClipboard(headers=False)
    qtable.deleteValues()


def paste_data_tab_separated(viewer: TableViewerBase):
    """Paste from tab separated text"""
    if table := _utils.get_mutable_table(viewer, None):
        table._qwidget.pasteFromClipBoard(sep="\t")
        return None

    import pandas as pd

    viewer.add_table(pd.read_clipboard(header=None, sep="\t"))
    return None


def paste_data_comma_separated(viewer: TableViewerBase):
    """Paste from comma separated text"""
    _utils.get_mutable_table(viewer)._qwidget.pasteFromClipBoard(sep=",")


def paste_data_from_numpy_string(viewer: TableViewerBase):
    """Paste from numpy-style text"""
    # import re
    import numpy as np
    import pandas as pd

    # TODO: use regex
    # repr_pattern = re.compile(r"array\(.*\)")
    # str_pattern = re.compile(r"\[.*\]")

    table = _utils.get_mutable_table(viewer)._qwidget
    s = _utils.get_clipboard_text().strip()

    _is_repr = s.startswith("array(") and s.endswith(")")
    _is_str = s.startswith("[") and s.endswith("]")

    if _is_repr:
        arr = eval(f"np.{s}", {"np": np, "__builtins__": {}}, {})
        if not isinstance(arr, np.ndarray):
            raise ValueError("Invalid numpy array representation.")
        if arr.ndim > 2:
            raise ValueError("Cannot paste array with dimension > 2.")
        return table._paste_data(pd.DataFrame(arr))
    elif _is_str:
        arr = np.asarray(eval(s.replace(" ", ", "), {"__builtins__": {}}, {}))
        if arr.ndim > 2:
            raise ValueError("Cannot paste array with dimension > 2.")
        return table._paste_data(pd.DataFrame(arr))
    else:
        raise ValueError("Invalid numpy array representation.")


def paste_data_from_markdown(viewer: TableViewerBase):
    """Paste from Markdown text"""
    import re
    import pandas as pd

    table = _utils.get_mutable_table(viewer)
    text = _utils.get_clipboard_text().strip()
    pattern = re.compile(r"(?=\|)(.+?)(?=\|)")
    lines = text.split("\n")

    # check the second "|---|---..." line.
    grid_spec = lines[1]
    if not re.match(r"\|(([:-]+\|)+)+", grid_spec):
        raise ValueError("Input text is not a markdown table.")

    # parse header
    columns = [cell.group().lstrip("|").strip() for cell in pattern.finditer(lines[0])]
    if len(columns) == 0:
        raise ValueError(f"Informal header: {lines[0]!r}")

    data: list[list[str]] = []
    for line in lines[2:]:
        if line := line.strip():
            data.append(
                [cell.group().lstrip("|").strip() for cell in pattern.finditer(line)]
            )

    df = pd.DataFrame(data, columns=columns, dtype="string")
    table.native._paste_data(df)
    return None


def paste_data_from_rst(viewer: TableViewerBase):
    """Paste from reStructuredText (rst) table"""

    import numpy as np
    import pandas as pd
    from docutils.parsers.rst.tableparser import GridTableParser, SimpleTableParser
    from docutils.statemachine import StringList
    from tabulous._pd_index import char_arange

    table = _utils.get_mutable_table(viewer)
    text = _utils.get_clipboard_text().strip()
    lines = list(filter(bool, (line.strip() for line in text.splitlines())))
    if lines[0].startswith("+"):
        parser = GridTableParser()
    else:
        parser = SimpleTableParser()

    colspec, header, data = parser.parse(StringList(lines))

    # TODO: multiple header is not supported
    if header:
        columns = []
        for rowspan, colspan, lineno, stringlist in header[-1]:
            name: str = "\n".join(stringlist.data)
            columns.append(name)
            for i in range(colspan):
                columns.append(f"{name}_{i}")
    else:
        colcount = 0
        for rowspan, colspan, lineno, stringlist in data[0]:
            colcount += 1 + colspan
        columns = char_arange(colcount)

    df = pd.DataFrame(
        np.empty((len(data), len(columns))), columns=columns, dtype="string"
    )
    ir = 0
    for row in data:
        ic = 0
        for cell in row:
            # (0, 0, 3, StringList(['column 3'], items=[(None, 3)]))
            rowspan, colspan, lineno, stringlist = cell
            value: str = "\n".join(stringlist.data)
            df.iloc[ir : ir + rowspan + 1, ic : ic + colspan + 1] = value
            ic += colspan + 1
        ir += 1

    table.native._paste_data(df)
    return None


def delete_values(viewer: TableViewerBase):
    """Delete selected cells"""
    _utils.get_mutable_table(viewer)._qwidget.deleteValues()


def add_highlight(viewer: TableViewerBase):
    """Add highlight to cells"""
    qwidget = _utils.get_table(viewer)._qwidget
    qwidget.setHighlights(qwidget.highlights() + qwidget.selections(map=True))


def delete_selected_highlight(viewer: TableViewerBase):
    """Delete selected highlight"""
    table = _utils.get_table(viewer)._qwidget
    table._qtable_view._highlight_model.delete_selected()
    table._qtable_view._selection_model.set_ctrl(False)


def show_context_menu(viewer: TableViewerBase):
    """Execute context menu"""
    qtable = _utils.get_table(viewer)._qwidget
    qtable.showContextMenuAtIndex()


def raise_slot_error(viewer: TableViewerBase):
    """Show traceback at the cell"""
    qtable = _utils.get_table(viewer)._qwidget
    qtable.raiseSlotError()


def _notify_editability(viewer: TableViewerBase):
    """Notify that current table is not editable."""
    viewer._qwidget._tablestack.notifyEditability()


def set_column_dtype(viewer: TableViewerBase):
    """Set column specific dtype for data conversion and validation."""
    from tabulous._dtype import QDtypeWidget

    sheet = _utils.get_spreadsheet(viewer)._qwidget
    col = _utils.get_selected_column(viewer)
    if out := QDtypeWidget.requestValue(sheet):
        dtype_str, validation, formatting = out
        if dtype_str == "unset":
            dtype_str = None
        colname = sheet._data_raw.columns[col]
        sheet.setColumnDtype(colname, dtype_str)
        if validation:
            sheet._set_default_data_validator(colname)
        if formatting:
            sheet._set_default_text_formatter(colname)
    return None


def insert_row_above(viewer: TableViewerBase):
    """Insert a row above"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    row, _ = sheet.current_index
    return sheet.native.insertRows(row, 1)


def insert_row_below(viewer: TableViewerBase):
    """Insert a row below"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    row, _ = sheet.current_index
    return sheet.native.insertRows(row + 1, 1)


def insert_column_left(viewer: TableViewerBase):
    """Insert a column left"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    _, col = sheet.current_index
    return sheet.native.insertColumns(col, 1)


def insert_column_right(viewer: TableViewerBase):
    """Insert a column right"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    _, col = sheet.current_index
    return sheet.native.insertColumns(col + 1, 1)


def remove_selected_rows(viewer: TableViewerBase):
    """Remove selected rows"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    row, col = sheet.current_index
    _, rng = sheet.native._qtable_view._selection_model.range_under_index(row, col)
    if rng is not None:
        row_range = rng[0]
        sheet.native.removeRows(row_range.start, row_range.stop - row_range.start)
        return None
    raise ValueError("No selection under cursor.")


def remove_selected_columns(viewer: TableViewerBase):
    """Remove selected columns"""
    sheet = _utils.get_spreadsheet(viewer)
    if not sheet.editable:
        return _notify_editability()
    row, col = sheet.current_index
    _, rng = sheet.native._qtable_view._selection_model.range_under_index(row, col)
    if rng is not None:
        col_range = rng[1]
        sheet.native.removeColumns(col_range.start, col_range.stop - col_range.start)
        return None
    raise ValueError("No selection under cursor.")


def write_data_signal_in_console(viewer: TableViewerBase):
    """Write data signal connection to console"""
    from qtpy import QtCore

    table = _utils.get_table(viewer)
    sels = table.selections
    if len(sels) != 1:
        return
    qviewer = viewer.native
    console = qviewer._console_widget
    if console is None or not console.isActive():
        delay = 500  # need delay to wait for the console to be activated
    else:
        delay = 0
    qviewer.setConsoleVisible(True)

    def _update_console():
        console = qviewer._console_widget
        rsl, csl = sels[0]
        if rsl == slice(None) and csl == slice(None):
            _getitem = ""
        else:
            r0 = str(rsl.start) if rsl.start is not None else ""
            r1 = str(rsl.stop) if rsl.stop is not None else ""
            c0 = str(csl.start) if csl.start is not None else ""
            c1 = str(csl.stop) if csl.stop is not None else ""
            _getitem = f"[{r0}:{r1}, {c0}:{c1}]"
        text = (
            f"@viewer.current_table.events.data{_getitem}.connect\n"
            "def _on_data_changed(info: 'ItemInfo'):\n"
            "    "
        )
        if buf := console.buffer():
            if not buf.endswith("\n"):
                buf += "\n"
            text = buf + text
        console.setBuffer(text)
        console.setFocus()
        console.setTempText("...")

    QtCore.QTimer.singleShot(delay, _update_console)
    return None


def write_slice_in_console(viewer: TableViewerBase) -> None:
    """Write data slice to console"""
    from qtpy import QtCore

    table = _utils.get_table(viewer)
    sels = table.selections
    if len(sels) != 1:
        return
    qviewer = viewer.native
    console = qviewer._console_widget
    if console is None or not console.isActive():
        delay = 500  # need delay to wait for the console to be activated
    else:
        delay = 0
    qviewer.setConsoleVisible(True)

    def _update_console():
        console = qviewer._console_widget
        rsl, csl = sels[0]
        if rsl.start is None:
            rsl_str = f"slice({rsl.stop})"
        elif rsl.stop is None:
            rsl_str = f"slice({rsl.start}, None)"
        else:
            rsl_str = f"slice({rsl.start}, {rsl.stop})"
        if csl.start is None:
            csl_str = f"slice({csl.stop})"
        elif csl.stop is None:
            csl_str = f"slice({csl.start}, None)"
        else:
            csl_str = f"slice({csl.start}, {csl.stop})"

        text = f"sl = ({rsl_str}, {csl_str})\n"
        if buf := console.buffer():
            if not buf.endswith("\n"):
                buf += "\n"
            text = buf + text
        console.setBuffer(text)
        console.setFocus()

    QtCore.QTimer.singleShot(delay, _update_console)
    return None


def write_data_reference_in_console(viewer: TableViewerBase) -> None:
    """Write data reference to console"""
    from qtpy import QtCore

    table = _utils.get_table(viewer)
    sels = table.selections
    if len(sels) != 1:
        return
    qviewer = viewer.native
    console = qviewer._console_widget
    if console is None or not console.isActive():
        delay = 500  # need delay to wait for the console to be activated
    else:
        delay = 0
    qviewer.setConsoleVisible(True)

    def _update_console():
        console = qviewer._console_widget
        rsl, csl = sels[0]
        r0 = rsl.start if rsl.start is not None else ""
        r1 = rsl.stop if rsl.stop is not None else ""
        c0 = csl.start if csl.start is not None else ""
        c1 = csl.stop if csl.stop is not None else ""

        text = f"viewer.data.iloc[{r0}:{r1}, {c0}:{c1}]"
        console.insertText(text)
        console.setFocus()

    QtCore.QTimer.singleShot(delay, _update_console)
    return None


def add_spinbox(viewer: TableViewerBase) -> None:
    """Add Spinbox"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.spinbox(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.SpinBox(**kwargs))
    return None


def add_float_spinbox(viewer: TableViewerBase) -> None:
    """Add FloatSpinBox"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.float_spinbox(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.FloatSpinBox(**kwargs))
    return None


def add_slider(viewer: TableViewerBase) -> None:
    """Add Slider"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.slider(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.Slider(**kwargs))
    return None


def add_float_slider(viewer: TableViewerBase) -> None:
    """Add FloatSlider"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.float_slider(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.FloatSlider(**kwargs))
    return None


def add_checkbox(viewer: TableViewerBase) -> None:
    """Add CheckBox"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.checkbox(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.CheckBox(**kwargs))
    return None


def add_radio_button(viewer: TableViewerBase) -> None:
    """Add RadioButton"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    kwargs = _dialogs.radio_button(parent=viewer.native)
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.RadioButton(**kwargs))
    return None


def add_line_edit(viewer: TableViewerBase) -> None:
    """Add LineEdit"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, mwdg.LineEdit())
    return None


def remove_cell_widgets(viewer: TableViewerBase) -> None:
    """Remove cell widgets"""
    sheet = _utils.get_spreadsheet(viewer)._qwidget
    with sheet._mgr.merging():
        for r, c in sheet._qtable_view._selection_model.iter_all_indices():
            sheet._set_widget_at_index(r, c, None)
    return None


def edit_current(viewer: TableViewerBase) -> None:
    """Edit current cell"""
    table = _utils.get_table(viewer)._qwidget
    table._qtable_view._edit_current()
    return None


def sort_by_columns(viewer: TableViewerBase) -> None:
    """Sort by column(s)"""
    table = _utils.get_table(viewer)
    indices = _utils.get_selected_columns(viewer)
    by = [table.columns[index] for index in indices]
    return table.proxy.sort(by=by)


def filter_by_columns(viewer: TableViewerBase) -> None:
    """Filter by a column"""
    table = _utils.get_table(viewer)
    indices = _utils.get_selected_columns(viewer)
    by = [table.columns[index] for index in indices]
    table.proxy.add_filter_buttons(by, show_menu=True)
    return None


def shuffle_data_column_wise(viewer: TableViewerBase) -> None:
    """Shuffle table data columnwise"""
    table = _utils.get_mutable_table(viewer)
    for sel in table.selections:
        selected_data = table._qwidget.dataShown(parse=False).iloc[sel]
        shuffled = selected_data.sample(frac=1)
        shuffled.reset_index()
        table.cell[sel] = shuffled
    return None


def sort_inplace(viewer: TableViewerBase) -> None:
    """Sort table data inplace"""
    table = _utils.get_mutable_table(viewer)
    sel = _utils.get_a_selection(table)
    df = table.data.iloc[sel]
    if df.shape[1] == 1:
        out = df.sort_values(by=df.columns[0])
    else:
        chosen = _dialogs.choose_multiple(
            choices={
                "choices": list(df.columns),
                "widget_type": ToggleSwitchSelect,
                "label": "by",
            },
            parent=viewer.native,
        )
        if chosen:
            out = df.sort_values(by=chosen)
        else:
            return None
    table.native.setDataFrameValue(*sel, out)
    return None



from __future__ import annotations

from qtpy.sip import isdeleted
from ._base import TableComponent


def _plt_function(name: str, method_name: str | None = None):
    """Make a function that calls ``plt.<name>`` on the current figure."""
    if method_name is None:
        method_name = name

    def func(self: PlotInterface, *args, **kwargs):
        ax = self.gca()
        out = getattr(ax, method_name)(*args, picker=True, **kwargs)
        self.draw()
        return out

    func.__name__ = name
    func.__qualname__ = f"PlotInterface.{name}"
    func.__doc__ = f"Call ``plt.{name}`` on the current figure."
    return func


class PlotInterface(TableComponent):
    """The interface of plotting."""

    def __init__(self, parent=TableComponent._no_ref):
        super().__init__(parent)
        self._current_widget = None

    def gcf(self):
        """Get current figure."""
        return self.gcw().figure

    def gca(self):
        """Get current axis."""
        return self.gcw().ax

    def gcw(self):
        """Get current widget."""
        if self._current_widget is None or isdeleted(self._current_widget):
            self.new_widget()
        return self._current_widget

    def clf(self):
        """Clear the current figure."""
        self.gcf().clf()
        return self.draw()

    def cla(self):
        """Clear the current axis."""
        self.gca().cla()
        return self.draw()

    def new_widget(self, nrows: int = 1, ncols: int = 1, style: str | None = None):
        """Create a new plot widget and add it to the table."""
        from tabulous._qt._plot import QtMplPlotCanvas

        table = self.parent
        qviewer = table._qwidget._qtable_view.parentViewer()

        if not qviewer._white_background and style is None:
            style = "dark_background"

        wdt = QtMplPlotCanvas(nrows=nrows, ncols=ncols, style=style)
        wdt.set_background_color(qviewer.backgroundColor().name())
        wdt.canvas.deleteRequested.connect(self.delete_widget)
        table.add_side_widget(wdt, name="Plot")
        self._current_widget = wdt
        return wdt

    def delete_widget(self) -> None:
        """Delete the current widget from the side area."""
        if self._current_widget is None:
            return None
        try:
            self.parent._qwidget._side_area.removeWidget(self._current_widget)
        except Exception:
            pass
        self._current_widget.deleteLater()
        self._current_widget = None
        return None

    def figure(self, style=None):
        return self.subplots(style=style)[0]

    def subplots(self, nrows=1, ncols=1, style=None):
        wdt = self.new_widget(nrows=nrows, ncols=ncols, style=style)
        return wdt.figure, wdt.axes

    plot = _plt_function("plot")
    plot_date = _plt_function("plot_date")
    quiver = _plt_function("quiver")
    scatter = _plt_function("scatter")
    bar = _plt_function("bar")
    errorbar = _plt_function("errorbar")
    hist = _plt_function("hist")
    text = _plt_function("text")
    fill_between = _plt_function("fill_between")
    fill_betweenx = _plt_function("fill_betweenx")

    def xlabel(self, *args, **kwargs):
        """Call ``plt.xlabel`` on the current side figure."""
        if not args and not kwargs:
            return self.gca().get_xlabel()
        out = self.gca().set_xlabel(*args, **kwargs)
        self.draw()
        return out

    def ylabel(self, *args, **kwargs):
        """Call ``plt.ylabel`` on the current side figure."""
        if not args and not kwargs:
            return self.gca().get_ylabel()
        out = self.gca().set_ylabel(*args, **kwargs)
        self.draw()
        return out

    def xlim(self, *args, **kwargs):
        """Call ``plt.xlim`` on the current side figure."""
        if not args and not kwargs:
            return self.gca().get_xlim()
        out = self.gca().set_xlim(*args, **kwargs)
        self.draw()
        return out

    def ylim(self, *args, **kwargs):
        """Call ``plt.ylim`` on the current side figure."""
        if not args and not kwargs:
            return self.gca().get_ylim()
        out = self.gca().set_ylim(*args, **kwargs)
        self.draw()
        return out

    def title(self, *args, **kwargs):
        """Call ``plt.title`` on the current side figure."""
        if not args and not kwargs:
            return self.gca().get_title()
        out = self.gca().set_title(*args, **kwargs)
        self.draw()
        return out

    def draw(self):
        """Update the current side figure."""
        return self._current_widget.draw()

    @property
    def background_color(self):
        """Background color of the current figure."""
        return self.gcf().get_facecolor()

    @background_color.setter
    def background_color(self, color):
        """Set background color of the current figure."""
        return self.gcw().set_background_color(color)



from __future__ import annotations
from pathlib import Path
from typing import Callable, TYPE_CHECKING, Union
from functools import partial
import weakref
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QAction

from tabulous._qt._svg import QColoredSVGIcon
from tabulous._qt._multitips import QHasToolTip
from tabulous import commands as cmds

from ._toolbutton import QColoredToolButton, QMoreToolButton
from ._corner import QSelectionRangeEdit

if TYPE_CHECKING:
    from tabulous._qt._mainwindow import _QtMainWidgetBase
    from tabulous.widgets._mainwindow import TableViewerBase

ICON_DIR = Path(__file__).parent.parent / "_icons"


class QSubToolBar(QtW.QToolBar, QHasToolTip):
    """The child toolbar widget."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._buttons: list[QColoredToolButton] = []

    def updateIconColor(self, color):
        """Update all the icons with the given color."""
        for button in self._buttons:
            button.updateColor(color)

    def appendAction(self, f: Callable, qicon: QColoredSVGIcon):
        """Add function ``f`` to the toolbar with the given icon."""
        btn = QColoredToolButton()
        self.addWidget(btn)
        btn.setIcon(qicon)
        btn.clicked.connect(f)
        if isinstance(f, partial):
            doc = f.func.__doc__
        else:
            doc = f.__doc__
        btn.setToolTip(doc)
        self._buttons.append(btn)
        return

    def appendMenuAction(self, f: Callable, name: str):
        btn = self._buttons[-1]
        if not isinstance(btn, QMoreToolButton):
            btn = QMoreToolButton()

            self.addWidget(btn)
            self._buttons.append(btn)

        menu = btn.menu()
        if menu is None:
            menu = QtW.QMenu(self)
            btn.setMenu(menu)

        action = menu.addAction(name)
        action.triggered.connect(f)
        if isinstance(f, partial):
            doc = f.func.__doc__
        else:
            doc = f.__doc__
        action.setToolTip(doc)
        return

    def toolTipPosition(self, index: int) -> QtCore.QPoint:
        btn = self._buttons[index]
        pos = btn.pos()
        pos.setY(pos.y() + btn.height() // 2)
        return pos

    def toolTipCount(self) -> int:
        return len(self._buttons)

    def clickButton(self, index: int, *, ignore_index_error: bool = True):
        """Emulate a click on the button at the given index."""
        if index < 0 or index >= len(self._buttons):
            if ignore_index_error:
                return
            else:
                raise IndexError("Index out of range")
        btn = self._buttons[index]
        return btn.click()


class QTableStackToolBar(QtW.QToolBar, QHasToolTip):
    _child_widgets: weakref.WeakValueDictionary[str, QSubToolBar]
    sliceChanged = Signal(object)

    def __init__(self, parent: _QtMainWidgetBase):
        super().__init__(parent)

        self._tab = QtW.QTabWidget(self)
        self._tab.setTabsClosable(False)
        self._tab.setContentsMargins(0, 0, 0, 0)

        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Minimum
        )
        self._child_widgets = weakref.WeakValueDictionary()

        self.addWidget(self._tab)
        self.setMaximumHeight(84)
        self.initToolbar()

    @property
    def _corner_widget(self) -> QSelectionRangeEdit:
        if self._tab.cornerWidget() is None:
            corner = QSelectionRangeEdit(self._tab)
            corner.sliceChanged.connect(self.sliceChanged.emit)

            self._tab.setCornerWidget(corner)
            corner.editingFinished.connect(lambda: self.parent().setCellFocus())
            corner.hide()
        return self._tab.cornerWidget()

    @property
    def viewer(self) -> TableViewerBase:
        """The parent viewer object."""
        return self.parent()._table_viewer

    def currentIndex(self) -> int:
        """Current tab index."""
        return self._tab.currentIndex()

    def setCurrentIndex(self, index: int):
        """Set current tab index."""
        return self._tab.setCurrentIndex(index)

    def currentToolBar(self) -> QSubToolBar:
        """Current toolbar."""
        return self._child_widgets[self._tab.tabText(self._tab.currentIndex())]

    def toolTipPosition(self, index: int) -> QtCore.QPoint:
        return self._tab.tabBar().tabRect(index).topLeft()

    def toolTipText(self, index: int) -> str:
        return list(self._child_widgets.keys())[index][0]

    def toolTipCount(self) -> int:
        return self._tab.count()

    # fmt: off
    if TYPE_CHECKING:
        def parent(self) -> _QtMainWidgetBase: ...
    # fmt: on

    def addToolBar(self, name: str):
        """Add a tab of toolbar of name ``name``."""
        if name in self._child_widgets.keys():
            raise ValueError(f"Tab with name {name!r} already exists.")
        toolbar = QSubToolBar(self)
        toolbar.setContentsMargins(0, 0, 0, 0)
        self._tab.addTab(toolbar, name)
        self._child_widgets[name] = toolbar

    def registerAction(self, tabname: str, f: Callable, icon: Union[str, Path]):
        """Register a callback `f` in tab `tabname`."""
        if tabname not in self._child_widgets:
            self.addToolBar(tabname)
        toolbar = self._child_widgets[tabname]
        qicon = QColoredSVGIcon.fromfile(icon)

        def fn():
            return f(self.viewer)

        fn.__doc__ = f.__doc__
        toolbar.appendAction(fn, qicon)
        return

    def addSeparatorToChild(self, tabname: str) -> QAction:
        toolbar = self._child_widgets[tabname]
        return toolbar.addSeparator()

    def registerMenuAction(self, tabname: str, f: Callable, name: str | None = None):
        """Register a menu in tab `tabname`."""
        toolbar = self._child_widgets[tabname]
        if name is None:
            name = f.__name__.replace("_", " ").capitalize()

        def fn():
            return f(self.viewer)

        fn.__doc__ = f.__doc__
        toolbar.appendMenuAction(fn, name)
        return None

    def setToolButtonColor(self, color: str):
        """Update all the tool button colors."""
        for toolbar in self._child_widgets.values():
            toolbar.updateIconColor(color)

    # fmt: off
    def initToolbar(self):
        """Add tool buttons"""

        main = self.parent()._is_mainwindow

        self.registerAction("Home", cmds.file.open_table, ICON_DIR / "open_table.svg")  # noqa: E501
        self.registerAction("Home", cmds.file.open_spreadsheet, ICON_DIR / "open_spreadsheet.svg")  # noqa: E501
        self.registerAction("Home", cmds.file.save_table, ICON_DIR / "save_table.svg")  # noqa: E501
        self.addSeparatorToChild("Home")
        self.registerAction("Home", cmds.file.open_sample, ICON_DIR / "open_sample.svg")  # noqa: E501
        self.addSeparatorToChild("Home")
        self.registerAction("Home", cmds.window.toggle_console, ICON_DIR / "toggle_console.svg")  # noqa: E501
        self.registerAction("Home", cmds.window.show_command_palette, ICON_DIR / "palette.svg")  # noqa: E501
        if main:
            self.addSeparatorToChild("Home")
            self.registerAction("Home", cmds.window.show_preference, ICON_DIR / "preferences.svg")  # noqa: E501

        self.registerAction("Edit", cmds.selection.copy_data_tab_separated, ICON_DIR / "copy.svg")  # noqa: E501
        self.registerAction("Edit", cmds.selection.paste_data_tab_separated, ICON_DIR / "paste.svg")  # noqa: E501
        self.registerAction("Edit", cmds.selection.cut_data, ICON_DIR / "cut.svg")  # noqa: E501
        self.addSeparatorToChild("Edit")
        self.registerAction("Edit", cmds.table.undo_table, ICON_DIR / "undo.svg")  # noqa: E501
        self.registerAction("Edit", cmds.table.redo_table, ICON_DIR / "redo.svg")  # noqa: E501

        self.registerAction("Table", cmds.table.copy_as_table, ICON_DIR / "copy_as_table.svg")  # noqa: E501
        self.registerAction("Table", cmds.table.copy_as_spreadsheet, ICON_DIR / "copy_as_spreadsheet.svg")  # noqa: E501
        self.addSeparatorToChild("Table")
        self.registerAction("Table", cmds.column.run_groupby, ICON_DIR / "groupby.svg")  # noqa: E501
        self.registerAction("Table", cmds.table.switch_columns, ICON_DIR / "switch_header.svg")  # noqa: E501
        self.registerAction("Table", cmds.table.pivot, ICON_DIR / "pivot.svg")
        self.registerAction("Table", cmds.table.melt, ICON_DIR / "melt.svg")
        self.addSeparatorToChild("Table")
        self.registerAction("Table", cmds.table.random, ICON_DIR / "random.svg")  # noqa: E501
        self.registerAction("Table", cmds.table.round, ICON_DIR / "round.svg")  # noqa: E501

        self.registerAction("Analyze", cmds.analysis.summarize_table, ICON_DIR / "summarize_table.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.analysis.show_eval_widget, ICON_DIR / "eval.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.table.show_finder_widget, ICON_DIR / "find_item.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.selection.sort_by_columns, ICON_DIR / "sort_table.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.analysis.show_filter_widget, ICON_DIR / "filter.svg")  # noqa: E501
        self.addSeparatorToChild("Analyze")
        self.registerAction("Analyze", cmds.analysis.show_optimizer_widget, ICON_DIR / "optimize.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.analysis.show_stats_widget, ICON_DIR / "stats_test.svg")  # noqa: E501
        self.registerAction("Analyze", cmds.analysis.show_sklearn_widget, ICON_DIR / "sklearn_analysis.svg")  # noqa: E501

        self.registerAction("View", cmds.view.set_popup_mode, ICON_DIR / "view_popup.svg")  # noqa: E501
        self.registerAction("View", cmds.view.set_dual_h_mode, ICON_DIR / "view_dual_h.svg")  # noqa: E501
        self.registerAction("View", cmds.view.set_dual_v_mode, ICON_DIR / "view_dual_v.svg")  # noqa: E501
        self.registerAction("View", cmds.view.reset_view_mode, ICON_DIR / "view_reset.svg")  # noqa: E501
        self.addSeparatorToChild("View")
        self.registerAction("View", cmds.tab.tile_tables, ICON_DIR / "tile.svg")  # noqa: E501
        self.registerAction("View", cmds.tab.untile_table, ICON_DIR / "untile.svg")  # noqa: E501
        self.addSeparatorToChild("View")
        self.registerAction("View", cmds.table.switch_layout, ICON_DIR / "switch_layout.svg")  # noqa: E501

        self.registerAction("Plot", cmds.plot.plot, ICON_DIR / "plot.svg")
        self.registerAction("Plot", cmds.plot.scatter, ICON_DIR / "scatter.svg")  # noqa: E501
        self.registerAction("Plot", cmds.plot.hist, ICON_DIR / "hist.svg")
        self.registerMenuAction("Plot", cmds.plot.bar, name="Run plt.bar")
        self.registerMenuAction("Plot", cmds.plot.errorbar, name="Run plt.errorbar")  # noqa: E501
        self.registerMenuAction("Plot", cmds.plot.fill_between, name="Run plt.fill_between")  # noqa: E501
        self.registerMenuAction("Plot", cmds.plot.fill_betweenx, name="Run plt.fill_betweenx")  # noqa: E501
        self.addSeparatorToChild("Plot")
        self.registerAction("Plot", cmds.plot.swarmplot, ICON_DIR / "swarmplot.svg")  # noqa: E501
        self.registerAction("Plot", cmds.plot.barplot, ICON_DIR / "barplot.svg")  # noqa: E501
        self.registerAction("Plot", cmds.plot.boxplot, ICON_DIR / "boxplot.svg")  # noqa: E501
        self.registerMenuAction("Plot", cmds.plot.boxenplot, name="Run sns.boxenplot")  # noqa: E501
        self.registerMenuAction("Plot", cmds.plot.stripplot, name="Run sns.stripplot")  # noqa: E501
        self.registerMenuAction("Plot", cmds.plot.violinplot, name="Run sns.violinplot")  # noqa: E501
        self.addSeparatorToChild("Plot")
        self.registerAction("Plot", cmds.plot.new_figure, ICON_DIR / "new_figure.svg")  # noqa: E501

        return
    # fmt: on



from __future__ import annotations

from dataclasses import dataclass, asdict
from functools import lru_cache
import json
from pathlib import Path
from tabulous.color import normalize_color


@dataclass(frozen=True)
class Style:
    background: str
    foreground: str
    base_color: str
    highlight0: str
    highlight1: str
    background0: str
    background1: str
    cell_highlight: str
    cell_selection: str
    inv_color: str

    @lru_cache(maxsize=12)
    def format_text(self, text: str) -> str:
        for name, value in asdict(self).items():
            text = text.replace(f"#[{name}]", f"{value}")
        return text

    def format_file(self, file: str | Path | None = None) -> str:
        if file is None:
            file = Path(__file__).parent / "_style.qss"
        with open(file) as f:
            text = f.read()
        return self.format_text(text)

    @classmethod
    def from_global(cls, name: str) -> Style:
        theme = GLOBAL_STYLES.get(name, None)
        if theme is None:
            raise ValueError(f"Theme {name!r} not found")
        js = asdict(theme)
        self = cls(**js)
        return self


GLOBAL_STYLES: dict[str, Style] = {}

with open(Path(__file__).parent / "defaults.json") as f:
    js: dict = json.load(f)
    for name, style in js.items():

        bg = normalize_color(style["background"])
        fg = normalize_color(style["foreground"])
        base = normalize_color(style["base_color"])
        if "background0" not in style:
            style["background0"] = bg.mix(fg, 0.1).html
        if "background1" not in style:
            style["background1"] = bg.mix(fg, -0.1).html
        if "highlight0" not in style:
            style["highlight0"] = base.mix(bg, 0.6).html
        if "highlight1" not in style:
            style["highlight1"] = base.mix(bg, 0.75).html
        if "cell_selection" not in style:
            style["cell_selection"] = base.html
        GLOBAL_STYLES[name] = Style(**style)



from __future__ import annotations

from types import MappingProxyType
from typing import Any, TYPE_CHECKING
from dataclasses import asdict, dataclass, field
from functools import wraps
from pathlib import Path
from contextlib import contextmanager
from appdirs import user_config_dir

if TYPE_CHECKING:
    KeyBinding = dict[str, str | list[str]]

TXT_PATH = Path(user_config_dir("tabulous", "tabulous", "history.txt"))
CONFIG_PATH = Path(user_config_dir("tabulous", "tabulous", "config.toml"))
CELL_NAMESPACE_PATH = Path(user_config_dir("tabulous", "tabulous", "cell_namespace.py"))
POST_INIT_PATH = Path(user_config_dir("tabulous", "tabulous", "post_init.py"))


def warn_on_exc(default=None):
    def _wrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except Exception as e:
                import warnings

                warnings.warn(f"{type(e).__name__}: {e}", UserWarning)
                return default

        return wrapper

    return _wrapper


@warn_on_exc(default=None)
def dump_file_open_path(path: str):
    """Save the recently opened file paths to a text file."""
    path = str(path)
    if not TXT_PATH.exists():
        TXT_PATH.parent.mkdir(parents=True, exist_ok=True)
        TXT_PATH.write_text("")
    with open(TXT_PATH, "r+") as f:
        lines = [s.strip() for s in f.readlines()]
        if path in lines:
            lines.remove(path)
            lines.append(path)
        else:
            lines.append(path)
            if len(lines) > 16:
                lines = lines[-16:]
        f.truncate(0)
        f.seek(0)
        f.write("\n".join(lines))

    return None


@warn_on_exc(default=[])
def load_file_open_path() -> list[str]:
    """Load the recently opened file paths to a text file."""
    if not TXT_PATH.exists():
        return []
    with open(TXT_PATH) as f:
        lines = [s.strip() for s in f.readlines()]

    return lines


def _compile_file(path: Path, default_text: str = ""):
    # if file doesn't exist, create it
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text("")
    with open(path, encoding="utf-8") as f:
        code_str = f.read()
        code = compile(code_str, path, "exec")
    if code_str.strip() == "":
        path.write_text(default_text.strip() + "\n")
    return code


@warn_on_exc(default=MappingProxyType({}))
def load_cell_namespace() -> MappingProxyType:
    """Load the cell namespace from a file."""
    code = _compile_file(CELL_NAMESPACE_PATH, default_text=_CELL_NAMESPACE_TEXT)
    ns: dict[str, Any] = {}
    exec(code, {}, ns)
    to_be_deleted = set()
    for k in ns.keys():
        if k.startswith("_"):
            to_be_deleted.add(k)

    if "__all__" in to_be_deleted:
        _all = ns.get("__all__", [])
        to_be_deleted.update(_all)
    for k in to_be_deleted:
        ns.pop(k, None)
    return MappingProxyType(ns)


def get_post_initializers():
    code = _compile_file(POST_INIT_PATH, default_text=_POST_INIT_TEXT)
    ns: dict[str, Any] = {}
    exec(code, {}, ns)

    if len(ns) == 0:
        # if file is empty, don't do anything
        return None

    from tabulous.post_init import TableInitializer, ViewerInitializer

    table_initializer = TableInitializer()
    viewer_initializer = ViewerInitializer()
    for var in ns.values():
        if isinstance(var, TableInitializer):
            table_initializer.join(var)
        elif isinstance(var, ViewerInitializer):
            viewer_initializer.join(var)
    return viewer_initializer, table_initializer


def prep_default_keybindings() -> KeyBinding:
    from tabulous.commands import DEFAULT_KEYBINDING_SETTING

    kb = {}
    for cmd, seq in DEFAULT_KEYBINDING_SETTING:
        mod = cmd.__module__.split(".")[-1]
        name = cmd.__name__
        kb[f"{mod}.{name}"] = seq
    return kb


def set_default_keybindings(kb: KeyBinding) -> KeyBinding:
    from tabulous.commands import DEFAULT_KEYBINDING_SETTING

    existing = set()
    for seq in kb.values():
        if isinstance(seq, str):
            existing.add(seq)
        elif isinstance(seq, list):
            existing.update(seq)
        else:
            raise RuntimeError(f"Invalid keybinding sequence: {seq}")

    for cmd, seq in DEFAULT_KEYBINDING_SETTING:
        if isinstance(seq, str):
            if seq not in existing:
                mod = cmd.__module__.split(".")[-1]
                name = cmd.__name__
                kb[f"{mod}.{name}"] = seq

        else:
            if not existing.intersection(seq):
                mod = cmd.__module__.split(".")[-1]
                name = cmd.__name__
                kb[f"{mod}.{name}"] = seq

    return kb


@dataclass
class ConsoleNamespace:
    """Default namespace of the console."""

    tabulous: str = "tbl"
    viewer: str = "viewer"
    pandas: str = "pd"
    numpy: str = "np"
    load_startup_file: bool = True


@dataclass
class Table:
    """Table settings."""

    max_row_count: int = 400000
    max_column_count: int = 40000
    font: str = "Arial"
    font_size: int = 10
    row_size: int = 28
    column_size: int = 100


@dataclass
class Cell:
    """Cell settings."""

    eval_prefix: str = "="
    ref_prefix: str = "&="


@dataclass
class Window:
    """Window config."""

    ask_on_close: bool = True
    show_console: bool = False
    theme: str = "light-blue"
    notify_latest: bool = True
    selection_editor: bool = True
    nonmain_style: bool = False
    animate: bool = True
    title_bar: str = "native"


@dataclass
class TabulousConfig:
    """The config model."""

    console_namespace: ConsoleNamespace = field(default_factory=ConsoleNamespace)
    table: Table = field(default_factory=Table)
    cell: Cell = field(default_factory=Cell)
    window: Window = field(default_factory=Window)
    keybindings: KeyBinding = field(default_factory=prep_default_keybindings)

    @classmethod
    def from_toml(cls, path: Path = CONFIG_PATH) -> TabulousConfig:
        """Load the config file."""
        import toml

        if not path.exists():
            return cls()

        with open(path) as f:
            tm = toml.load(f)
        return cls.from_dict(tm)

    @classmethod
    def from_dict(cls, dict_: dict[str, Any]) -> TabulousConfig:
        console_namespace = dict_.get("console_namespace", {})
        table = dict_.get("table", {})
        cell = dict_.get("cell", {})
        window = dict_.get("window", {})
        kb = set_default_keybindings(dict_.get("keybindings", {}))
        return cls(
            ConsoleNamespace(**_as_fields(console_namespace, ConsoleNamespace)),
            Table(**_as_fields(table, Table)),
            Cell(**_as_fields(cell, Cell)),
            Window(**_as_fields(window, Window)),
            kb,
        )

    def as_toml(self):
        """Save the config file."""
        import toml

        if not CONFIG_PATH.exists():
            CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
            CONFIG_PATH.write_text("")
        with open(CONFIG_PATH, "w") as f:
            toml.dump(asdict(self), f)

        return None

    def as_immutable(self) -> MappingProxyType:
        """Return the immutable version of the config."""
        return MappingProxyType(
            {k: MappingProxyType(v) for k, v in asdict(self).items()}
        )


CONFIG: TabulousConfig | None = None


def get_config(reload: bool = False) -> TabulousConfig:
    """Get the global config."""
    global CONFIG
    if CONFIG is None or reload:
        CONFIG = TabulousConfig.from_toml()
    return CONFIG


def update_config(cfg: TabulousConfig, save: bool = False) -> None:
    """Update the global config."""
    global CONFIG
    CONFIG = cfg
    if save:
        CONFIG.as_toml()
    return None


@contextmanager
def init_config():
    """Temporarily initialize the config."""
    global CONFIG
    if CONFIG is None:
        get_config()
    ori = TabulousConfig()
    old = CONFIG
    CONFIG = ori
    try:
        yield CONFIG
    finally:
        CONFIG = old
        old.as_toml()


def _as_fields(kwargs: dict[str, Any], dcls: type) -> dict[str, Any]:
    """Remove dict keys that does not belong to the dataclass fields."""
    fields = set(dcls.__annotations__.keys())
    return {k: v for k, v in kwargs.items() if k in fields}


_CELL_NAMESPACE_TEXT = """
# File for custom namespace in the table cells.

# Variables defined in the `__all__` list will be available in the cells.
# By, uncommenting following lines, you can use such as
# `=SUM(df.iloc[:, 0])` in cells.

# import scipy
# import numpy as np

# __all__ = ["scipy", "SUM", "AVERAGE"]

# def SUM(x):
#     return np.sum(x)

# def AVERAGE(x):
#     return np.mean(x)
"""

_POST_INIT_TEXT = """
# File for post-initialization of the viewer and table.
# You can add custom actions, commands and variables on the startup of the
# application.

# First, get initializer objects.

# from tabulous.post_init import get_initializers
# viewer, table = get_initializers()

# These objects has similar interface to the viewer and table classes.

# 1. Add custom actions to the right-click context menu of table columns.

# @table.columns.register("User defined > Print column name")
# def _print_column_name(table, index):
#     print(table.columns[index])

# 2. Add custom keybindings to viewers and tables.

# @viewer.keymap.register("Ctrl+K, Ctrl+1")
# def _my_keybinding(viewer):
#     print(viewer)

# @table.keymap.register("Ctrl+Shift+K, Ctrl+1")
# def _my_keybinding(table):
#     print(table)

# 3. Add custom variables to the console (note that you can also update
#    the IPython startup files to do the same thing).

# viewer.console.update({"PI": 3.141592653589793})

# 4. Add custom variables to the table cells (not that you can also update
#    the cell_namespace.py file to do the same thing).

# import numpy as np
# viewer.cell_namespace.update({"SUM": np.sum})

"""



from __future__ import annotations
from qtpy import QtWidgets as QtW
from . import _utils


class QAbstractEval(QtW.QWidget):
    _BUTTON_TEXT = ""
    _PLACEHOLDER_TEXT = ""

    def __init__(self, parent: QtW.QWidget | None = None):
        super().__init__(parent)
        self._label = QtW.QLabel(">>>", self)
        self._line = _utils.QCompletableLineEdit(self)
        self._btn = QtW.QPushButton(self._BUTTON_TEXT)
        self._btn.clicked.connect(self.callback)
        self._line.enterClicked.connect(self.callback)
        self._line.setPlaceholderText(self._PLACEHOLDER_TEXT)

        _layout = QtW.QHBoxLayout()
        _layout.addWidget(self._label)
        _layout.addWidget(self._line)
        _layout.addWidget(self._btn)
        self.setLayout(_layout)

        self.setToolTip(self.__class__.__doc__.replace("\n    ", "\n").strip())

    def callback(self):
        raise NotImplementedError()

    def lineEdit(self) -> QtW.QLineEdit:
        return self._line


class QLiteralEvalWidget(QAbstractEval):
    """
    Evaluate literal string.

    >>> result = val * 3  # Update or create "result" column
    >>> val_norm = val - val.mean()  # Use DataFrame method
    """

    _BUTTON_TEXT = "Eval"
    _PLACEHOLDER_TEXT = "e.g. result = val * 3"

    def callback(self):
        """Evaluate the current text as a Python expression."""
        text = self._line.text()
        if text == "":
            return
        table = self._line.currentPyTable(assert_mutable=True)
        table.query(text)
        self._line.toHistory()


class QLiteralFilterWidget(QAbstractEval):
    """
    Apply filter using literal string.

    Examples
    --------
    >>> variable < 1.4  # values in "variable" column is < 1.4
    >>> a < b  # values in "a" column < values in "b" column
    >>> None  # reset filter
    """

    _BUTTON_TEXT = "Filter"
    _PLACEHOLDER_TEXT = "e.g. length < 3.5"

    def callback(self):
        """Update the filter of the current table using the expression."""
        text = self._line.text()
        if text == "":
            return
        table = self._line.currentPyTable()
        table.proxy.filter(text)
        self._line.toHistory()



from __future__ import annotations
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt


class QHasToolTip(QtW.QWidget):
    """A trait for a widget that has multiple tooltips."""

    _labels: list[QKeyComboTip] = []

    def toolTipPosition(self, index: int) -> QtCore.QPoint:
        """Return the position of the tooltip for the given index."""
        raise NotImplementedError()

    def toolTipText(self, index: int) -> str:
        """Return the text of the tooltip for the given index."""
        if 0 <= index < 9:
            return str(index + 1)
        elif index == 9:
            return "0"
        else:
            raise ValueError("Index must be between 0 and 9.")

    def toolTipCount(self) -> int:
        """Return the number of tooltips."""
        raise NotImplementedError()

    def showTabTooltips(self):
        """Show all the tooltips."""
        self._labels = []
        num = min(self.toolTipCount(), 10)
        for i in range(num):
            label = QKeyComboTip(self.toolTipText(i), self)
            pos = self.toolTipPosition(i)
            label.move(self.mapToGlobal(pos))
            label.show()
            self._labels.append(label)

    def hideTabTooltips(self):
        """Hide all the tooltips."""
        for label in self._labels:
            label.hide()
        self._labels = []


class QKeyComboTip(QtW.QLabel):
    def __init__(self, text: str, parent=None, size=14):
        super().__init__(text, parent, Qt.WindowType.ToolTip)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        font = self.font()
        font.setPointSize(size)
        self.setFont(font)
        self.setFixedSize(size + 4, size + 4)
        self.setStyleSheet("border: 1px solid gray;")
        self.hide()



from __future__ import annotations
from typing import TYPE_CHECKING
from . import _dialogs, _utils
from tabulous._magicgui import ToggleSwitchSelect
from tabulous.widgets._source import Source

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase, TableBase, SpreadSheet
    from ._arange import _RangeDialog


def new_spreadsheet(viewer: TableViewerBase):
    """Add an empty spreadsheet."""
    viewer.add_spreadsheet()
    viewer.native.setCellFocus()


def copy_as_table(viewer: TableViewerBase):
    """Copy current table as a new table"""
    table = _utils.get_table(viewer)
    out = viewer.add_table(table.data, name=f"{table.name}-copy")
    out._source = Source.from_table(table)


def copy_as_spreadsheet(viewer: TableViewerBase):
    """Copy current table as a new spreadsheet"""
    table = _utils.get_table(viewer)
    out = viewer.add_spreadsheet(table.data, name=f"{table.name}-copy")
    out._source = Source.from_table(table)


def copy_to_clipboard(viewer: TableViewerBase):
    table = _utils.get_table(viewer)
    table._qwidget.dataShown().to_clipboard()
    return None


def switch_index(viewer: TableViewerBase):
    """Switch index header and the left column"""
    table = _utils.get_mutable_table(viewer)
    table._qwidget._switch_head_and_index(axis=0)


def switch_columns(viewer: TableViewerBase):
    """Switch column header and the top row"""
    table = _utils.get_mutable_table(viewer)
    table._qwidget._switch_head_and_index(axis=1)


def concat(viewer: TableViewerBase):
    """Concatenate table data (pd.concat)"""
    if len(viewer.tables) < 2:
        raise ValueError("At least two tables are required.")
    out = _dialogs.concat(
        viewer={"bind": viewer},
        names={
            "value": [viewer.current_table.name],
            "widget_type": ToggleSwitchSelect,
            "choices": [t.name for t in viewer.tables],
        },
        axis={"choices": [("vertical", 0), ("horizontal", 1)]},
        parent=viewer._qwidget,
    )
    if out is not None:
        table = viewer.add_table(out, name="concat")
        table._source = Source.from_table(viewer.current_table)


def merge(viewer: TableViewerBase):
    """Merge two tables (pd.merge)"""

    def _update_choices(wdt):
        table0: TableBase = wdt.merge.value
        table1: TableBase = wdt.with_.value
        if table0 is None or table1 is None:
            return
        col0 = table0.columns
        col1 = table1.columns
        choices = list(set(col0) & set(col1))
        wdt.on.choices = choices

    out = _dialogs.merge(
        merge={"changed": _update_choices},
        with_={"changed": _update_choices},
        how={"choices": ["left", "right", "outer", "inner"], "value": "inner"},
        on={"choices": [], "widget_type": ToggleSwitchSelect},
        parent=viewer._qwidget,
    )
    if out is not None:
        viewer.add_table(out, name="merged")


def pivot(viewer: TableViewerBase):
    """Pivot current table data (pd.pivot)"""
    table = _utils.get_table(viewer)
    col = list(table.data.columns)
    if len(col) < 3:
        raise ValueError("Table must have at least three columns.")
    out = _dialogs.pivot(
        df={"bind": table.data},
        index={"choices": col, "value": col[0]},
        columns={"choices": col, "value": col[1]},
        values={"choices": col, "value": col[2]},
        parent=viewer._qwidget,
    )
    if out is not None:
        table_out = viewer.add_table(out, name=f"{table.name}-pivot")
        table_out._source = Source.from_table(table)


def melt(viewer: TableViewerBase):
    """Melt (unpivot) current table data (pd.melt)"""
    table = _utils.get_table(viewer)
    cols = _utils.get_selected_columns(viewer)
    df = table.data
    out = df.melt(id_vars=[df.columns[i] for i in cols])
    if out is not None:
        table_out = viewer.add_table(out, name=f"{table.name}-melt")
        table_out._source = Source.from_table(table)


def transpose(viewer: TableViewerBase):
    """Transpose current table data"""
    table = _utils.get_table(viewer)
    out = table.data.T
    table_out = viewer.add_table(out, name=f"{table.name}-transposed")
    table_out._source = Source.from_table(table)


def fillna(viewer: TableViewerBase):
    """Fill nan values (pd.fillna)"""
    table = _utils.get_table(viewer)

    def _cb(mgui):
        mgui.value.visible = mgui.method.value == "value"

    _choices = [
        ("fill by value", "value"),
        ("forward fill", "ffill"),
        ("backward fill", "bfill"),
    ]
    out = _dialogs.fillna(
        df={"bind": table.data},
        method={
            "choices": _choices,
            "value": "value",
            "nullable": False,
            "changed": _cb,
        },
        value={"widget_type": "LiteralEvalLineEdit", "value": "0.0"},
        parent=viewer._qwidget,
    )
    if out is not None:
        table_out = viewer.add_table(out, name=f"{table.name}-fillna")
        table_out._source = Source.from_table(table)


def dropna(viewer: TableViewerBase):
    """Drop nan values (pd.dropna)"""
    table = _utils.get_table(viewer)
    out = _dialogs.dropna(
        df={"bind": table.data},
        axis={"choices": [("drop rows", 0), ("drop columns", 1)]},
        how={"choices": [("if any nan", "any"), ("if all nan", "all")]},
        parent=viewer._qwidget,
    )
    if out is not None:
        table_out = viewer.add_table(out, name=f"{table.name}-dropna")
        table_out._source = Source.from_table(table)


def show_finder_widget(viewer: TableViewerBase):
    """Toggle finder widget"""
    return viewer._qwidget._tablestack.openFinderDialog()


def reset_proxy(viewer: TableViewerBase) -> None:
    """Reset proxy (sort/filter)"""
    table = _utils.get_table(viewer)
    table.proxy.reset()
    return None


def random(viewer: TableViewerBase):
    """Generate random values"""
    table = viewer.current_table
    if table is None:
        return
    from ._random_data import RandomGeneratorDialog

    dlg = RandomGeneratorDialog()
    dlg.native.setParent(viewer._qwidget, dlg.native.windowFlags())
    dlg._selection_wdt._read_selection(table)
    dlg.show()

    @dlg.called.connect
    def _on_called():
        val = dlg.get_value(table._qwidget.model().df)
        rsl, csl, data = val
        table.cell[rsl, csl] = data


def round(viewer: TableViewerBase):
    """Round table data"""
    table = _utils.get_mutable_table(viewer)
    from magicgui.widgets import request_values

    out = request_values({"decimals": int}, parent=viewer._qwidget)
    if out is None:
        return
    decimals: int = out["decimals"]
    for sel in table.selections:
        selected_data = table.data_shown.iloc[sel]
        table.cell[sel] = selected_data.round(decimals)
    return None


def _run_range_dialog(dlg: _RangeDialog, viewer: TableViewerBase, table: TableBase):
    dlg.native.setParent(viewer._qwidget, dlg.native.windowFlags())
    dlg._selection._read_selection(table)
    dlg.show()

    @dlg.called.connect
    def _on_called():
        val = dlg.get_value(table._qwidget.model().df)
        rsl, csl, data = val
        c0 = csl.start
        was_empty = table.columns.size <= c0
        table.cell[rsl, c0] = data
        if was_empty and table.table_type == "SpreadSheet":
            _table: SpreadSheet = table
            _table.dtypes[_table.columns[c0]] = data.dtype


def date_range(viewer: TableViewerBase):
    """Generate a range of date values (pd.date_range)"""
    from ._arange import DateRangeDialog

    table = _utils.get_mutable_table(viewer)
    _run_range_dialog(DateRangeDialog(), viewer, table)


def timedelta_range(viewer: TableViewerBase):
    """Generate a range of timedelta values (pd.timedelta_range)"""
    from ._arange import TimeDeltaRangeDialog

    table = _utils.get_mutable_table(viewer)
    _run_range_dialog(TimeDeltaRangeDialog(), viewer, table)


def interval_range(viewer: TableViewerBase):
    """Generate a range of interval values (pd.interval_range)"""
    from ._arange import IntervalRangeDialog

    table = _utils.get_mutable_table(viewer)
    _run_range_dialog(IntervalRangeDialog(), viewer, table)


def period_range(viewer: TableViewerBase):
    """Generate a range of period values (pd.period_range)"""
    from ._arange import PeriodRangeDialog

    table = _utils.get_mutable_table(viewer)
    _run_range_dialog(PeriodRangeDialog(), viewer, table)


def toggle_editability(viewer: TableViewerBase):
    """Toggle table editability"""
    table = viewer.current_table
    try:
        table.editable = not table.editable
    except Exception:
        pass
    viewer.native.setCellFocus()


def show_undo_stack_view(viewer: TableViewerBase):
    """Show undo stack view"""
    _utils.get_table(viewer)._qwidget.undoStackView()


def undo_table(viewer: TableViewerBase):
    """Undo table operation"""
    _utils.get_mutable_table(viewer)._qwidget.undo()


def redo_table(viewer: TableViewerBase):
    """Redo table operation"""
    _utils.get_mutable_table(viewer)._qwidget.redo()


def switch_layout(viewer: TableViewerBase):
    """Switch table layout"""
    table = _utils.get_table(viewer)
    if table.layout == "vertical":
        table.layout = "horizontal"
    else:
        table.layout = "vertical"


def resize_cells_to_contents(viewer: TableViewerBase):
    """Resize cells to contents"""
    table = _utils.get_table(viewer)
    table.native._qtable_view.resizeColumnsToContents()
    table.native._qtable_view.resizeRowsToContents()


def zoom_in(viewer: TableViewerBase):
    """Zoom in"""
    table = _utils.get_table(viewer)
    table.native._qtable_view.zoomIn(1)


def zoom_out(viewer: TableViewerBase):
    """Zoom in"""
    table = _utils.get_table(viewer)
    table.native._qtable_view.zoomIn(-1)


def jump_to_cell(viewer: TableViewerBase):
    """Jump to a cell."""
    from tabulous._qt._jump import QJumpWidget

    if viewer.current_table is None:
        return
    jw = QJumpWidget(viewer.native)
    jw.show()



from __future__ import annotations

import builtins
from functools import wraps
import ast
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Sequence,
    overload,
    Any,
    Callable,
)

import numpy as np
from tabulous.types import ProxyType, _IntArray, _BoolArray, _IntOrBoolArray
from ._base import TableComponent
from tabulous._sort_filter_proxy import ComposableFilter, ComposableSorter

if TYPE_CHECKING:
    import pandas as pd
    from tabulous.widgets._table import TableBase


class ProxyInterface(TableComponent):
    """
    Interface to the table sorting/filtering.

    To sort table, use ``sort`` method.

    >>> table.proxy.sort('A', ascending=True)  # sort by column "A"
    >>> table.proxy.sort(['A', 'B'], ascending=True)  # sort by columns "A" and "B"

    To filter table, use ``filter`` method.

    >>> table.proxy.filter('A > 0')  # filter by column "A"
    >>> table.proxy.filter('(A > 0) & (B < 0)')  # filter by columns "A" and "B"

    To use custom proxy function, use ``set`` method. A proxy function must return
    a boolean array or an integer array of the same length as the number of rows.

    >>> table.proxy.set(my_filter_func)
    >>> @table.proxy.set
    >>> def my_filter_func(df: pd.DataFrame):
    ...     # do something

    To initialize the proxy, use ``reset`` method.

    >>> table.proxy.reset()

    """

    # fmt: off
    @overload
    def sort(self, by: str | Sequence[str], ascending: bool = True, compose: bool = False) -> None: ...  # noqa: E501
    @overload
    def sort(self, func: Callable[[pd.DataFrame], _IntArray]) -> None: ...
    # fmt: on

    def sort(self, by, ascending: bool = True, compose: bool = False) -> None:
        """
        Apply sort proxy to the table.

        If column names are given, sort button(s) will be added to the header.
        """
        table = self.parent

        if callable(by):
            if not ascending or compose:
                raise TypeError(
                    "Arguments 'ascending' and 'compose' are not supported to be "
                    "used with a callable input."
                )

            @wraps(by)
            def _sort(df: pd.DataFrame) -> _IntArray:
                arr = np.asarray(by(df))
                if arr.ndim != 1:
                    raise TypeError("The callable must return a 1D array.")
                elif arr.dtype.kind not in "ui":
                    raise TypeError("The callable must return an integer array.")
                return arr

            return table.proxy.set(_sort)

        from tabulous._qt._proxy_button import QHeaderSortButton

        if isinstance(by, str):
            by = [by]

        with table.undo_manager.merging(
            lambda cmds: f"table.proxy.sort(by={by!r}, ascending={ascending!r})"
        ):
            if not (compose and isinstance(table.proxy.obj, ComposableSorter)):
                table.proxy.reset()
            for x in by:
                index = table.columns.get_loc(x)
                QHeaderSortButton.install_to_table(
                    table.native, index, ascending=ascending
                )
        return None

    def add_sort_buttons(self, columns: str | list[str]) -> None:
        """Add sort buttons to the given column header sections."""
        return self.sort(by=columns, ascending=True, compose=False)

    # fmt: off
    @overload
    def filter(self, expr: str, namespace: dict = {}) -> None: ...
    @overload
    def filter(self, func: Callable[[pd.DataFrame], _BoolArray]) -> None: ...
    # fmt: on

    def filter(self, expr, namespace={}) -> None:
        """Apply filter proxy to the table."""
        if callable(expr):
            func = expr
            if namespace:
                raise TypeError("Cannot use a namespace with a callable.")

            def _filter(df: pd.DataFrame) -> _BoolArray:
                arr = np.asarray(func(df))
                if arr.ndim != 1:
                    raise TypeError("The callable must return a 1D array.")
                elif arr.dtype.kind != "b":
                    raise TypeError("The callable must return a boolean array.")
                return arr

        else:
            if _cfil := _try_ast_parse(expr, self.parent.columns):
                from tabulous._qt._proxy_button import QHeaderFilterButton

                index = _cfil.indices().pop()
                QHeaderFilterButton.install_to_table(self.parent.native, index)
                self.parent._qwidget._set_proxy(_cfil)
                return None

            def _filter(df: pd.DataFrame) -> np.ndarray:
                ns = dict(**dict(df.items()), **namespace)
                ns["__builtins__"] = _BUILTINS
                return eval(expr, ns, {})

            _filter.__name__ = f"filter<{expr!r}>"

        self.parent._qwidget.setProxy(_filter)
        return None

    def add_filter_buttons(
        self,
        columns: str | list[str],
        *,
        show_menu: bool = False,
    ) -> None:
        """Add filter buttons to the given column header sections."""
        from tabulous._qt._proxy_button import QHeaderFilterButton

        table = self.parent
        if isinstance(columns, str):
            columns = [columns]
        else:
            columns = list(columns)

        with table.undo_manager.merging(
            lambda cmds: f"table.proxy.add_filter_buttons({columns!r})"
        ):
            if not isinstance(table.proxy.obj, ComposableFilter):
                table.proxy.reset()
            for x in reversed(columns):
                index = table.columns.get_loc(x)
                btn = QHeaderFilterButton.install_to_table(table.native, index)
        if show_menu:
            btn.showMenu()
        return None

    def hide_buttons(
        self,
        columns: str | list[str] | None = None,
        missing_ok: bool = False,
    ) -> None:
        """Hide buttons in header sections."""
        if columns is None:
            # hide all buttons
            self.parent.native.updateHorizontalHeaderWidget({})
            return None

        if isinstance(columns, str):
            indices = [self.parent.columns.get_loc(columns)]
        else:
            indices = [self.parent.columns.get_loc(c) for c in columns]
        wdts = self.parent.native._header_widgets().copy()
        popped = [wdts.pop(index, None) for index in indices]

        if not missing_ok and None in popped:
            if isinstance(columns, str):
                raise ValueError(f"Column {columns!r} does not have a header button.")
            else:
                colnames = {columns[i] for i, x in popped if x is None}
                raise ValueError(f"Columns {colnames!r} do not have header buttons.")

        return None

    def reset(self) -> None:
        """Reset filter or sort."""
        return self._set_value(None)

    def set(self, proxy: ProxyType, check_duplicate: bool = True) -> None:
        """Set filter or sort."""
        if isinstance(proxy, (list, tuple, set)):
            proxy = np.asarray(proxy)
        if check_duplicate:
            if callable(proxy):

                @wraps(proxy)
                def _wrapped(df: pd.DataFrame) -> ProxyType:
                    out = proxy(df)
                    return _check_duplicate(out)

                _proxy = _wrapped
            elif proxy is None:
                _proxy = None
            else:
                _proxy = _check_duplicate(np.asarray(proxy))
            return self._set_value(_proxy)
        return self._set_value(proxy)

    @contextmanager
    def released(self, keep_widgets: bool = False):
        """Release the proxy in this context."""
        proxy = self.obj
        qtable = self.parent.native
        if keep_widgets:
            widgets = qtable._header_widgets().copy()
        else:
            widgets = {}
        try:
            self._set_value(None)
            yield
        finally:
            self._set_value(proxy)
            if widgets:
                qtable.updateHorizontalHeaderWidget(widgets)

    def apply(self, df: pd.DataFrame) -> pd.DataFrame:
        """Apply the current proxy to the DataFrame."""
        return self._get_proxy_object().apply(df)

    def as_indexer(self) -> np.ndarray:
        """Return the indexer that represents the current proxy."""
        return self._get_proxy_object().as_indexer(self.parent.data)

    @property
    def proxy_type(self):
        """Return the current proxy type."""
        return self._get_proxy_object().proxy_type

    @property
    def obj(self) -> ProxyType | None:
        """The proxy object."""
        return self._get_proxy_object()._obj

    @property
    def is_ordered(self) -> bool:
        """Return True if the proxy does not change the order of rows"""
        return self._get_proxy_object().is_ordered

    def _set_value(self, value: Any):
        return self.parent._qwidget.setProxy(value)

    def __set__(self, obj: TableBase, value: ProxyType):
        return super().__set__(obj, value)

    def _get_proxy_object(self):
        """Return the current proxy function."""
        return self.parent._qwidget._proxy


def _check_duplicate(out: _IntOrBoolArray):
    if out.dtype.kind in "ui" and np.unique(out).size != out.size:
        raise ValueError(f"The proxy contains duplicates: {out!r}")
    return out


def _try_ast_parse(expr: str, columns: pd.Index) -> ComposableFilter | None:
    ast_body = ast.parse(expr, mode="eval").body
    if not isinstance(ast_body, ast.Compare):
        return None
    try:
        cfil = ComposableFilter.from_ast(ast_body, columns)
    except Exception:
        return None
    return cfil


# "safe" builtin functions
# fmt: off
_BUILTINS = {
    k: getattr(builtins, k)
    for k in [
        "int", "str", "float", "bool", "list", "tuple", "set", "dict", "range",
        "slice", "frozenset", "len", "abs", "min", "max", "sum", "any", "all",
        "divmod", "id", "bin", "oct", "hex", "hash", "iter", "isinstance",
        "issubclass", "ord"
    ]
}
# fmt: on



from __future__ import annotations
from enum import Enum


class Marker(Enum):
    """Marker shapes."""

    point = "."
    pixel = ","
    circle = "o"
    triangle_down = "v"
    triangle_up = "^"
    triangle_left = "<"
    triangle_right = ">"
    tri_down = "1"
    tri_up = "2"
    tri_left = "3"
    tri_right = "4"
    octagon = "8"
    square = "s"
    pentagon = "p"
    star = "*"
    hexagon1 = "h"
    hexagon2 = "H"
    plus = "+"
    x = "x"
    diamond = "D"
    thin_diamond = "d"
    vline = "|"
    hline = "-"
    none = "None"


class LineStyle(Enum):
    """Line styles."""

    solid = "-"
    dashed = "--"
    dotted = ":"
    dash_dot = "-."
    none = "None"


class HatchStyle(Enum):
    """Hatch styles."""

    diagonal = "/"
    backdiagonal = "\\"
    vertical = "|"
    horizontal = "-"
    crossed = "+"
    cross_diagonal = "x"
    small_circle = "o"
    large_circle = "O"
    dots = "."
    stars = "*"
    none = "None"


class VerticalAlignment(Enum):
    """Vertical alignment."""

    top = "top"
    center = "center"
    bottom = "bottom"
    baseline = "baseline"
    center_baseline = "center_baseline"


class HorizontalAlignment(Enum):
    """Horizontal alignment."""

    left = "left"
    center = "center"
    right = "right"



from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from ._base import ViewerComponent

if TYPE_CHECKING:
    from tabulous.widgets import TableViewerBase


class Toolbar(ViewerComponent):
    """The toolbar proxy."""

    @property
    def visible(self) -> bool:
        """Visibility of the toolbar."""
        return self.parent._qwidget.toolBarVisible()

    @visible.setter
    def visible(self, val) -> None:
        return self.parent._qwidget.setToolBarVisible(val)

    @property
    def current_index(self) -> int:
        return self.parent._qwidget._toolbar.currentIndex()

    @current_index.setter
    def current_index(self, val: int) -> None:
        return self.parent._qwidget._toolbar.setCurrentIndex(val)

    def register(self, loc: str, f: Callable):
        raise NotImplementedError()


class Console(ViewerComponent):
    """The QtConsole proxy."""

    @property
    def visible(self) -> bool:
        """Visibility of the console."""
        return self.parent._qwidget.consoleVisible()

    @visible.setter
    def visible(self, val) -> None:
        return self.parent._qwidget.setConsoleVisible(val)

    @property
    def is_active(self) -> bool:
        return self.parent._qwidget._console_widget is not None

    def _get_console_widget(self):
        console = self.parent._qwidget._console_widget
        if console is None:
            raise RuntimeError("Console is not active.")
        return console

    @property
    def buffer(self) -> str:
        """Return the current text buffer of the console."""
        return self._get_console_widget().input_buffer

    @buffer.setter
    def buffer(self, val) -> None:
        return self._get_console_widget().setBuffer(val)

    def execute(self):
        """Execute current buffer."""
        return self._get_console_widget().execute()

    def update(self, ns: dict[str, Any]):
        """Update IPython namespace."""
        console = self.parent._qwidget._console_widget
        if console is None:
            self.parent._qwidget._queued_ns.update(ns)
        else:
            self._get_console_widget().update_console(ns)
        return None


class CommandPalette(ViewerComponent):
    def register(
        self,
        name: str,
        command: Callable[[TableViewerBase], Any] | None = None,
        *,
        key: str | None = None,
    ):
        """Register a command."""
        from tabulous.commands import register_command

        if callable(name):
            name = None
            command = name

        elif ":" in name:
            title, desc = name.split(":", maxsplit=1)
            title = title.strip()
            desc = desc.strip()
        else:
            title = "User defined"
            desc = name

        if key is not None:
            self.parent.keymap.register(key, command)
        return register_command(command, title=title, desc=desc)

    @property
    def visible(self) -> bool:
        """Visibility of the command palette."""
        qviewer = self.parent._qwidget
        return qviewer._command_palette.get_widget(qviewer).isVisible()

    @visible.setter
    def visible(self, val) -> None:
        qviewer = self.parent._qwidget
        return qviewer._command_palette.get_widget(qviewer).setVisible(val)



from __future__ import annotations

from typing import Hashable, Iterator, TYPE_CHECKING, Literal, Union, SupportsIndex
from functools import singledispatch
import re

from tabulous.exceptions import UnreachableError

if TYPE_CHECKING:
    import pandas as pd
    from typing_extensions import Self

_Slice = Union[int, slice]
_SlicePattern = r"(\w|:|-|\s)"


class SelectionOperator:
    """An object that defines a selection on a dataframe."""

    args: tuple
    PATTERN: str

    def fmt(self, df_expr: str = "df") -> str:
        """Format selection literal for display."""
        raise NotImplementedError()

    def fmt_iloc(self, df: pd.DataFrame, df_expr: str = "df") -> str:
        """Format selection literal as iloc indices."""
        rsel, csel = self.as_iloc(df)
        return f"{df_expr}.iloc[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def fmt_scalar(self, df_expr: str = "df") -> str:
        """Format 1x1 selection literal as a scalar reference."""
        raise NotImplementedError()

    def resolve_indices(self, df: pd.DataFrame, shape: tuple[int, int]):
        rsel, csel = self.as_iloc(df)
        nr, nc = shape
        if rsel == slice(None):
            rsel = slice(nr)
        if csel == slice(None):
            csel = slice(nc)
        return self.from_iloc(rsel, csel, df)

    def __format__(self, spec: str) -> str:
        return self.fmt(df_expr=spec)

    def __repr__(self) -> str:
        cname = type(self).__name__
        return f"{cname}({self.fmt()})"

    def operate(self, df: pd.DataFrame) -> pd.DataFrame | pd.Series:
        """Slice dataframe according to selection literal."""
        raise NotImplementedError()

    def as_iloc(self, df: pd.DataFrame) -> tuple[_Slice, _Slice]:
        """Return selection literal as iloc indices."""
        raise NotImplementedError()

    def as_iloc_slices(self, df: pd.DataFrame) -> tuple[slice, slice]:
        """Return selection literal as iloc indices, forcing slices."""
        rsl, csl = self.as_iloc(df)
        if isinstance(rsl, SupportsIndex):
            r = rsl.__index__()
            rsl = slice(r, r + 1)
        elif rsl == slice(None):
            rsl = slice(0, df.index.size)
        if isinstance(csl, SupportsIndex):
            c = csl.__index__()
            csl = slice(c, c + 1)
        elif csl == slice(None):
            csl = slice(0, df.columns.size)
        return rsl, csl

    def as_iat(self, df: pd.DataFrame) -> tuple[int, int]:
        rsl, csl = self.as_iloc(df)
        if isinstance(rsl, SupportsIndex):
            r = rsl.__index__()
        elif isinstance(rsl, slice):
            if rsl.stop - rsl.start != 1:
                raise ValueError(f"{self} is not a scalar reference.")
            r = rsl.start
        if isinstance(csl, SupportsIndex):
            c = csl.__index__()
        elif isinstance(csl, slice):
            if csl.stop - csl.start != 1:
                raise ValueError(f"{self} is not a scalar reference.")
            c = csl.start
        return r, c

    def shape(self, df: pd.DataFrame) -> tuple[int, int]:
        """Return the shape of the selection."""
        rsl, csl = self.as_iloc_slices(df)
        return (rsl.stop - rsl.start, csl.stop - csl.start)

    def area(self, df: pd.DataFrame) -> int:
        """Return the number of cells selected."""
        rsl, csl = self.as_iloc_slices(df)
        return (rsl.stop - rsl.start) * (csl.stop - csl.start)

    @classmethod
    def from_iloc(cls, r: _Slice, c: _Slice, df: pd.DataFrame) -> Self:
        """Construct selection literal from iloc indices."""
        raise NotImplementedError()

    def __eq__(self, other: Self) -> bool:
        if isinstance(other, type(self)):
            return self.args == other.args
        return False


class ColumnSelOp(SelectionOperator):
    """An object that represents selection such as ``df["foo"][2:5]``."""

    args: tuple[Hashable, slice]
    PATTERN = r"df\[.+?\]\[.+?\]"

    def __init__(self, col: Hashable, rows: slice):
        self.args = (col, rows)

    def fmt(self, df_expr: str = "df") -> str:
        col, rows = self.args
        return f"{df_expr}[{_fmt_slice(col)}][{_fmt_slice(rows)}]"

    def fmt_scalar(self, df_expr: str = "df") -> str:
        col, rows = self.args
        start, stop = rows.start, rows.stop
        if stop - start != 1:
            raise ValueError("Cannot format as a scalar value.")
        return f"{df_expr}[{_fmt_slice(col)}][{_fmt_slice(start)}]"

    def operate(self, df: pd.DataFrame) -> pd.DataFrame:
        col, rows = self.args
        return df[col][rows]

    def as_iloc(self, df: pd.DataFrame) -> tuple[_Slice, _Slice]:
        colname, rsel = self.args
        col = df.columns.get_loc(colname)
        return rsel, col

    @classmethod
    def from_iloc(cls, r: _Slice, c: _Slice, df: pd.DataFrame) -> Self:
        if isinstance(c, slice):
            if c.start == c.stop - 1:
                c = c.start
            else:
                raise ValueError("Cannot convert slice to row selection.")
        colname = df.columns[c]
        return cls(colname, r)


class LocSelOp(SelectionOperator):
    """An object that represents selection such as ``df.loc["foo":"bar", 2:5]``."""

    PATTERN = r"df\.loc\[.+?\]"

    def __init__(self, rsel: Hashable | slice, csel: Hashable | slice):
        self.args = (rsel, csel)

    def fmt(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        return f"{df_expr}.loc[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def fmt_scalar(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        if isinstance(rsel, slice):
            if _has_none(rsel) or rsel.start != rsel.stop:
                raise ValueError("Cannot format as a scalar value.")
            rsel = rsel.start
        if isinstance(csel, slice):
            if _has_none(csel) or csel.start != csel.stop:
                raise ValueError("Cannot format as a scalar value.")
            csel = csel.start
        return f"{df_expr}.loc[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def operate(self, df: pd.DataFrame) -> pd.DataFrame:
        rsel, csel = self.args
        return df.loc[rsel, csel]

    def as_iloc(self, df: pd.DataFrame) -> tuple[_Slice, _Slice]:
        rsel, csel = self.args
        index, columns = df.axes
        if isinstance(rsel, slice):
            if rsel.start is None:
                rstart = None
            else:
                rstart = index.get_loc(rsel.start)
            if rsel.stop is None:
                rstop = None
            else:
                rstop = index.get_loc(rsel.stop) + 1
            irsel = slice(rstart, rstop)
        else:
            irsel = index.get_loc(rsel)
        if isinstance(csel, slice):
            if csel.start is None:
                cstart = None
            else:
                cstart = columns.get_loc(csel.start)
            if csel.stop is None:
                cstop = None
            else:
                cstop = columns.get_loc(csel.stop) + 1
            icsel = slice(cstart, cstop)
        else:
            icsel = columns.get_loc(csel)
        return irsel, icsel

    @classmethod
    def from_iloc(cls, r: _Slice, c: _Slice, df: pd.DataFrame) -> Self:
        """Construct operator from an iloc-style slices."""
        index, columns = df.axes
        if isinstance(r, slice):
            rsel = _normalize_loc_slice(r, index)
        else:
            rsel = index[r]
        if isinstance(c, slice):
            csel = _normalize_loc_slice(c, columns)
        else:
            csel = columns[c]
        return cls(rsel, csel)


def _normalize_loc_slice(sl: slice, index: pd.Index) -> slice:
    if sl.start is None:
        start = 0
    else:
        start = sl.start
    if sl.stop is None:
        stop = -1
    else:
        stop = sl.stop
    return slice(index[start], index[stop - 1])


class ILocSelOp(SelectionOperator):
    """An object that represents selection such as ``df.iloc[4:7, 2:5]``."""

    args: tuple[_Slice, _Slice]
    PATTERN = rf"df\.iloc\[{_SlicePattern}+(,{_SlicePattern}*)?\]"

    def __init__(self, rsel: _Slice, csel: _Slice):
        self.args = (rsel, csel)

    def fmt(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        return f"{df_expr}.iloc[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def fmt_scalar(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        if isinstance(rsel, slice):
            if _has_none(rsel) or rsel.start != rsel.stop - 1:
                raise ValueError(f"Cannot format {(rsel, csel)} as a scalar value.")
            rsel = rsel.start
        if isinstance(csel, slice):
            if _has_none(csel) or csel.start != csel.stop - 1:
                raise ValueError(f"Cannot format {(rsel, csel)} as a scalar value.")
            csel = csel.start
        return f"{df_expr}.iloc[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def operate(self, df: pd.DataFrame) -> pd.DataFrame | pd.Series:
        rsel, csel = self.args
        return df.iloc[rsel, csel]

    def as_iloc(self, df: pd.DataFrame = None) -> tuple[_Slice, _Slice]:
        return self.args

    def as_iloc_slices(self, df: pd.DataFrame | None = None) -> tuple[slice, slice]:
        return super().as_iloc_slices(df)

    @classmethod
    def from_iloc(cls, r: _Slice, c: _Slice, df: pd.DataFrame = None) -> Self:
        """Construct operator from an iloc-style slices."""
        return cls(r, c)


class ValueSelOp(SelectionOperator):
    """An object that represents selection such as ``df.iloc[4:7, 2:5]``."""

    PATTERN = rf"df\.values\[{_SlicePattern}+(,{_SlicePattern}*)?\]"

    def __init__(self, rsel: _Slice, csel: _Slice):
        self.args = (rsel, csel)

    def fmt(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        return f"{df_expr}.values[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def fmt_scalar(self, df_expr: str = "df") -> str:
        rsel, csel = self.args
        if isinstance(rsel, slice):
            if _has_none(rsel) or rsel.start != rsel.stop - 1:
                raise ValueError("Cannot format as a scalar value.")
            rsel = rsel.start
        if isinstance(csel, slice):
            if _has_none(csel) or csel.start != csel.stop - 1:
                raise ValueError("Cannot format as a scalar value.")
            csel = csel.start
        return f"{df_expr}.values[{_fmt_slice(rsel)}, {_fmt_slice(csel)}]"

    def operate(self, df: pd.DataFrame) -> pd.DataFrame:
        rsel, csel = self.args
        return df.values[rsel, csel]

    def as_iloc(self, df: pd.DataFrame) -> tuple[_Slice, _Slice]:
        return self.args

    @classmethod
    def from_iloc(cls, r: _Slice, c: _Slice, df: pd.DataFrame) -> Self:
        """Construct operator from an iloc-style slices."""
        return cls(r, c)


def iter_extract(text: str, *, df_expr: str = "df") -> Iterator[SelectionOperator]:
    """Iteratively extract selection literal from text."""
    for match_obj in _PATTERN.finditer(text):
        try:
            op = parse(match_obj.group(), df_expr=df_expr)
        except (ValueError, SyntaxError):
            pass
        else:
            yield op


def iter_extract_with_range(
    text: str,
    *,
    df_expr: str = "df",
) -> Iterator[tuple[tuple[int, int], SelectionOperator]]:
    for match_obj in _PATTERN.finditer(text):
        range = (match_obj.start(), match_obj.end())
        yield range, parse(match_obj.group(), df_expr=df_expr)


def parse(expr: str, *, df_expr: str = "df") -> SelectionOperator:
    """Parse dataframe-slicing expression."""
    ndf = len(df_expr)
    if expr.startswith(f"{df_expr}["):
        # df['val'][...]
        colname, rsl_str = _split_or(expr[ndf + 1 : -1], "][")
        rsl = _parse_slice(rsl_str)
        sel = ColumnSelOp(_eval(colname), rsl)

    elif expr.startswith(f"{df_expr}.loc["):
        # df.loc[..., ...]
        rsl_str, csl_str = _split_or(expr[ndf + 5 : -1], ",")
        rsl = _parse_slice(rsl_str)
        csl = _parse_slice(csl_str)
        sel = LocSelOp(rsl, csl)

    elif expr.startswith(f"{df_expr}.iloc["):
        # df.iloc[..., ...]
        rsl_str, csl_str = _split_or(expr[ndf + 6 : -1], ",")
        rsl = _parse_slice(rsl_str)
        csl = _parse_slice(csl_str)
        sel = ILocSelOp(rsl, csl)

    elif expr.startswith(f"{df_expr}.values["):
        # df.values[..., ...]
        rsl_str, csl_str = _split_or(expr[ndf + 8 : -1], ",")
        rsl = _parse_slice(rsl_str)
        csl = _parse_slice(csl_str)
        sel = ValueSelOp(rsl, csl)

    else:
        raise UnreachableError(f"{expr!r}")

    return sel


def construct(
    rsl: slice,
    csl: slice,
    df: pd.DataFrame,
    method: Literal["loc", "iloc", "values"] = "loc",
    column_selected: bool = False,
    allow_out_of_bounds: bool = False,
) -> SelectionOperator | None:
    """Construct a selection operator from given slices and data frame."""

    nr, nc = df.shape

    # normalize out-of-bound
    if not allow_out_of_bounds:
        rsl = slice(*rsl.indices(nr))
        if rsl.stop > nr:
            if rsl.start >= nr:
                return None
            rsl = slice(rsl.start, nr)
        csl = slice(*csl.indices(nc))
        if csl.stop > nc:
            if csl.start >= nc:
                return None
            csl = slice(csl.start, nc)

    if column_selected:
        rsl = slice(None)

    try:
        if method == "loc":
            if csl.start == csl.stop - 1:
                selop = ColumnSelOp.from_iloc(rsl, csl, df)
            else:
                selop = LocSelOp.from_iloc(rsl, csl, df)
        elif method == "iloc":
            selop = ILocSelOp.from_iloc(rsl, csl, df)
        elif method == "values":
            selop = ValueSelOp.from_iloc(rsl, csl, df)
        else:
            raise RuntimeError(f"Unknwon slicing mode {method!r}")
    except IndexError:
        # out-of-bound in spreadsheet
        return None
    return selop


_PATTERN = re.compile(
    "|".join(s.PATTERN for s in (ColumnSelOp, LocSelOp, ILocSelOp, ValueSelOp))
)


def find_all_dataframe_expr(s: str) -> list[str]:
    return _PATTERN.findall(s)


def find_last_dataframe_expr(s: str) -> int:
    """
    Detect last `df[...][...]` expression from given string.

    Returns start index if matched, otherwise -1.
    """
    _match = None
    for _match in _PATTERN.finditer(s):
        pass
    return _match.start() if _match else -1


def _eval(expr: str, default=None):
    """evaluate expression."""
    return eval(expr, {"__builtins__": {}}, {}) if expr else default


def _parse_slice(s: str) -> Hashable | slice:
    s = s.strip()
    if ":" in s:
        start_str, stop_str = s.split(":")
        start = _eval(start_str)
        stop = _eval(stop_str)
        return slice(start, stop)
    else:
        if s == "":
            return slice(None)
        return _eval(s)


def _has_none(sl: slice):
    return sl.start is None or sl.stop is None


@singledispatch
def _fmt_slice(s) -> str:
    return str(s)


@_fmt_slice.register
def _(s: int) -> str:
    return str(s)


@_fmt_slice.register
def _(s: str) -> str:
    return repr(s)


@_fmt_slice.register
def _(s: slice) -> str:
    if s == slice(None):
        return ":"
    start = "" if s.start is None else s.start
    stop = "" if s.stop is None else s.stop
    return f"{start!r}:{stop!r}"


def _split_or(s: str, sep: str, default: str = ":") -> tuple[str, str]:
    if sep not in s:
        return s, default
    return s.split(sep)
